#!/usr/bin/env python3
#
# Copyright (C) 2020 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

import setuptools

setuptools.setup()
